/*---------------------------
Vvoj okennho systmu ve WinAPI
Bakalsk prce 2006, 
autor: Dolej Petr
login: xdolej06
Fakulta Informanch Technologi v Brn
---------------------------*/
/*Samotn jdro IVVieweru*/

#include "ivview_functions.h"
#include "../Library/Common.h"
#include "../Okenni rozhrani/SoWindow.h"
#include "../Okenni rozhrani/SoWinMenu.h"

extern SoOrthographicCamera *overlayCamera;  // overlay workaround by PCJohn
extern SbBool useWalkViewer;
extern FileInfo files;
extern FileInfo files2;
extern SbBool showPerfMeter;
extern int numTris;
extern int numFramesRendered;
extern SbBool ivfix;
extern SbBool shapeHintBackface;
extern SoWinViewer *viewer;
extern SoWinMENU Menu;

//////////////////////////////////////////////////////////////////////
//
// Print the help message
//
 void
print_usage()
{
    fprintf(stderr, "Usage: %s [-hpw] [infiles]\n", progname);
    fprintf(stderr,
            "\t-h : Print this message (help) and exit\n"
            "\t-p : Enable performance meter\n"
            "\t-w : Use walk viewer (examiner viewer is default) - ignored because\n"
            "       Coin does not support SoWalkViewer yet\n"
            "If no input file name is given, stdin is used.\n"
            );

    exit(99);
}

/////////////////////////////////////////////////////////////////////
//
// Parse the command line arguments
//
 void
parse_args(int argc, char **argv)
{
    int err = FALSE;        // Flag: error in options?
    int c;

    showPerfMeter = FALSE;

    while ((c = getopt(argc, argv, "hpqw")) != -1) {
        switch(c) {
          case 'h':        // Help
            err = TRUE;
            break;
          case 'p':
            showPerfMeter = TRUE;
            break;
          case 'q':
            // no-op for backward compatibility only
            break;
          case 'w':
            useWalkViewer = TRUE;
            break;
          default:
            err = TRUE;
            break;
        }
    }

    // Can't handle -pw, at least right now:
    if (showPerfMeter && useWalkViewer) {
        fprintf(stderr, "Sorry, can't show performance when using "
                "walk viewer.  Ignoring -p option\n");
        showPerfMeter = FALSE;
    }

    /* Handle multiple filenames */
    if (optind == argc) {
        // Act like one argument "-" was given
        argc++;
        argv[optind] = "-";
    }

    files.count = argc - optind;
    files.names = (char **) malloc (files.count * sizeof(char *));
    for (int i=0; i<files.count; i++, optind++)
        files.names[i] = argv[optind];

    if (err)
        print_usage();
}


/////////////////////////////////////////////////////////////////////
//
// Read the scene from the file(s).
// If there are multiple files, combine their contents under one separator.
// Returns TRUE if read was OK, FALSE if read error.
//
 SbBool
readScene(FileInfo &files)
{
    SoInput in;
    SoSeparator *root;

    if (files.count == 0) // stdin is represented as "-" also if args
        return 0;         // are empty (parse_args will take care of it)

    // Watch out!  There's a hidden dependence here between this code
    // and the code in the optimizationChanged routine that runs files
    // through ivfix.  In particular, the optimizationChanged code
    // assumes that, if stdin is being read from, that no other
    // filenames are given.
    if (files.count==1 && strcmp(files.names[0], "-")==0) {
        
        if (!OPEN_INPUT_FILE(&in, files.names[0], FALSE, NULL))
            return 0;

        root = SoDB::readAll(&in);
        if (root == NULL)
            FILE_READ_ERROR(files.names[0], NULL, FALSE);
        else
            root->ref();

        CLOSE_INPUT_FILE(&in, files.names[0]);
    }
    else {          // Not reading from stdin:
        root = new SoSeparator;
        root->ref();

        //
        // Combine all the files under one separator
        //
        for (int i=0; i<files.count; i++) {
            
            if (!OPEN_INPUT_FILE(&in, files.names[i], FALSE, NULL)) {
                root->unref();
                return 0;
            }
            
            SoSeparator *inputRoot = SoDB::readAll(&in);
            if (inputRoot == NULL) {
                root->unref();
                FILE_READ_ERROR(files.names[i], NULL, FALSE);
                return 0;
            }
            root->addChild(inputRoot);
            
            CLOSE_INPUT_FILE(&in, files.names[i]);
        }
    }

    if (root && showPerfMeter) {
        SoSeparator *oldRoot = root;
        root = addUpdateCounter(oldRoot);
        root->ref();
        oldRoot->unref();
        numTris = countTriangles(root);
    }

    // Nuke old scene(s):
    if (files.original) {
        files.original->unref();
        files.original = NULL;
    }
    if (files.optimized) {
        files.optimized->unref();
        files.optimized = NULL;
    }
    if (files.ivfixed) {
        files.ivfixed->unref();
        files.ivfixed = NULL;
    }

    // And create new ones:
    if (root) {
        files.original = root;
        optimizationChanged(files, FALSE);
    }

    return (root != NULL);
}

/////////////////////////////////////////////////////////////////////
//
// Called when optimization options change, or when a new file is
// read.
//
 void
optimizationChanged(FileInfo &files, SbBool maintainCamera)
{
    // If no original file, have nothing to do:
    if (files.original == NULL) return;

    // We'll recreate the optimized graph:
    if (files.optimized) {
        files.optimized->unref();
        files.optimized = NULL;
    }
#if 0 // Not ported yet. Actually, this functionality is not accessible,
      // because the menu is not ported yet. PCJohn 2005-03-15
    if (ivfix) {
        SbBool err = FALSE;
        char *tempFile = NULL;
        setBusyCursor(TRUE);

        if (files.ivfixed == NULL) do {

            // First, special case-- read from stdin:
            if (files.count==1 && strcmp(files.names[0], "-")==0) {
                tempFile = tempnam(NULL, "ivfix");
                if (tempFile == NULL) {
                    err = 1;
                    break;
                }
                //
                // Run ivfix on the original scene
                //

                // String long enough for temp filename plus "ivfix >"
                char ivfixCommand[L_tmpnam + 10];
                sprintf(ivfixCommand, "ivfix > %s", tempFile);
                FILE *ivfixPipe = popen(ivfixCommand, "w");
                if (ivfixPipe == NULL) {
                    err = TRUE;
                    break;
                }

                // Write to ivfix through pipe:
                SoOutput out;
                out.setBinary(TRUE);
                out.setFilePointer(ivfixPipe);
                SoWriteAction wa(&out);
                wa.apply(files.original);
                if (ferror(ivfixPipe)) {
                    err = TRUE;
                    break;
                }
                pclose(ivfixPipe);

                // Now read tempFile:
                SoInput in;
                if (!in.openFile(tempFile)) {
                    err = TRUE;
                    break;
                }
                files.ivfixed = SoDB::readAll(&in);
                if (files.ivfixed == NULL) {
                    err = TRUE;
                    break;
                } else {
                    files.ivfixed->ref();
                }
                in.closeFile();
            }
            else {  // Not reading from stdin:
                files.ivfixed = new SoSeparator;
                files.ivfixed->ref();
                for (int i = 0; i < files.count; i++) {
                    const char *filename = files.names[i];
                    char *ivfixCommand = (char *)
                        alloca(strlen("ivfix ")+strlen(filename)+1);
                    sprintf(ivfixCommand, "ivfix %s", filename);

                    // Read from ivfix:
                    FILE *ivfixPipe = popen(ivfixCommand, "r");
                    if (ivfixPipe == NULL) {
                        files.ivfixed->unref();
                        files.ivfixed = NULL;
                        err = TRUE;
                        break;
                    }

                    // If the filename includes a directory path, add
                    // the directory name to the list of directories
                    // where to look for input files:
                    const char *slashPtr;
                    char *searchPath = NULL;
                    if ((slashPtr = strrchr(filename, '/')) != NULL) {
                        searchPath = strdup(filename);
                        searchPath[slashPtr - filename] = '\0';
                        SoInput::addDirectoryFirst(searchPath);
                    }

                    SoInput in;
                    in.setFilePointer(ivfixPipe);
                    SoSeparator *root = SoDB::readAll(&in);

                    if (searchPath) {
                        SoInput::removeDirectory(searchPath);
                        free(searchPath);
                    }

                    if (root == NULL) {
                        files.ivfixed->unref();
                        files.ivfixed = NULL;
                        err = TRUE;
                        break;
                    }
                    files.ivfixed->addChild(root);

                    pclose(ivfixPipe);
                }
            }

            if (showPerfMeter) {
                SoNode *oldRoot = files.ivfixed;
                files.ivfixed = addUpdateCounter(oldRoot);
                files.ivfixed->ref();
                oldRoot->unref();
                numTris = countTriangles(files.ivfixed);
            }
        } while (0);

        if (files.ivfixed) {
            files.optimized = files.ivfixed;
            files.optimized->ref();
        }

        setBusyCursor(FALSE);

        // Clean up temporary file (whether or not there was an error):
        if (tempFile && unlink(tempFile) < 0) {
            fprintf(stderr, "Warning: Error removing %s after ivfix\n",
                    tempFile);
        }
        if (tempFile) free(tempFile);

        // If there was an error:
        if (err) {
            system("xmessage 'Error running ivfix' > /dev/null");
        }
    }
#endif

    if (shapeHintBackface) {
        SoSeparator *optRoot = new SoSeparator;
        optRoot->renderCaching = SoSeparator::OFF;
        optRoot->boundingBoxCaching = SoSeparator::OFF;
        optRoot->pickCulling = SoSeparator::OFF;

        if (files.optimized != NULL) {
            // user also wants ivfix....
            optRoot->addChild(files.ivfixed);
            files.optimized->unref();
            files.optimized = NULL;
        } else {
            optRoot->addChild(files.original);
        }
        files.optimized = optRoot;
        files.optimized->ref();

        // Insert a ShapeHints node as the first child of files.optimized
        SoShapeHints *hints = new SoShapeHints;
        hints->vertexOrdering = SoShapeHints::COUNTERCLOCKWISE;
        hints->shapeType = SoShapeHints::SOLID;
        hints->setOverride(TRUE);
        files.optimized->insertChild(hints, 0);
    }

    //
    // Set scene graph to either optimized or original scene graph:
    //
    SoNode *currentCamera = NULL;
    if (maintainCamera) {
        SoNode *viewerCamera = viewer->getCamera();
        if (viewerCamera) {
            // Make a copy:
            currentCamera = viewerCamera->copy(FALSE);
            currentCamera->ref();
        }
    }

    //
    // Setting the scene graph
    // This code has been modified, because of using superscene (files.super)
    // as a workaround of overlays. PCJohn 2005-03-15
    ((SoGroup*)files.super->getChild(0))->removeAllChildren();
    if (files.optimized) {
        // optimized AND/OR ivfixed
        ((SoGroup*)files.super->getChild(0))->addChild(files.optimized);
        if (!currentCamera)
          viewer->viewAll();
    } else {
        ((SoGroup*)files.super->getChild(0))->addChild(files.original);
        if (!currentCamera)
          viewer->viewAll();
    }

    if (currentCamera) {
        SoNode *viewerCamera = viewer->getCamera();
        if (viewerCamera) {
            // Copy back:
            viewerCamera->copyFieldValues(currentCamera, FALSE);
        }
        currentCamera->unref();
    }
}

////////////////////////////////////////////////////////////////////
//
// Initialize the performance meter in the overlay planes
//
#ifdef _WIN32
 void
initPerfMeter(SoWinViewer *vwr)
#else
 void
initPerfMeter(SoQtViewer *vwr)
#endif
{
    //
    // Create the overlay scene
    //
    SoAnnotation *perfMeter = new SoAnnotation;

    // Add a callback node which will set the viewport
    SoCallback *overlayCB = new SoCallback;
    overlayCB->setCallback(overlayViewportCB);
    perfMeter->addChild(overlayCB);

    // Add an orthographic camera
    overlayCamera = new SoOrthographicCamera;
    perfMeter->addChild(overlayCamera);

    // Add the text string to display the tris per second
    char str[150];
    SoFont *perfMeterFont = new SoFont;
    perfMeterFont->size = 18;
    perfMeterFont->name = "Helvetica";
    perfMeter->addChild(perfMeterFont);
    SoText2 *perfMeterText = new SoText2;
    perfMeterText->justification = SoText2::LEFT;
    sprintf (str, "%d tris/frame", numTris);
    perfMeterText->string.setValue(str);
    perfMeter->addChild(perfMeterText);

#if 0 // This code is no longer used because overlays are not supported on
      // all platforms. Therefore, superscene workaround is used. PCJohn 2005-03-15
    SbColor col(1, 1, 1);
    vwr->setOverlayColorMap(1, 1, &col);
    vwr->setOverlaySceneGraph(perfMeter);
#endif
    ((SoGroup*)files.super->getChild(1))->removeAllChildren();
    ((SoGroup*)files.super->getChild(1))->addChild(perfMeter);

    // Update the performance meter every two seconds
    SoTimerSensor *perfSensor =
                new SoTimerSensor(perfSensorCB, perfMeterText);
    perfSensor->setInterval(2.0);
    perfSensor->schedule();

    // Reset the meter to only show the number of tris/frame -
    // otherwise the numbers might look misleading
    vwr->addStartCallback(viewStartCB, perfMeterText);
}


 void
overlayViewportCB(void *, SoAction *action)
{
#if 0 // Overlays is not used (see comments in previous function).
      // So, viewport is no longer needed. Instead, the correct
      // camera projection is set up. PCJohn 2005-03-15
    if (action->isOfType(SoGLRenderAction::getClassTypeId()))
        glViewport(0, 0, 40, 40);
#endif
    if (action->isOfType(SoGLRenderAction::getClassTypeId())) {
        const SbViewportRegion &vr = ((SoGLRenderAction*)action)->getViewportRegion();
        float pointSize = overlayCamera->height.getValue() / vr.getWindowSize()[1];
        float posX = pointSize * vr.getWindowSize()[0] / 2.f - pointSize * 20.f;
        float posY = overlayCamera->height.getValue() * 0.5f - pointSize * 20.f;
        overlayCamera->position.setValue(posX, posY, 1.f);
    }
}


////////////////////////////////////////////////////////////////////
//
// A sensor callback in the overlay scene updates the
// performance meter while the object is animating.
//
 void
perfSensorCB(void *data, SoSensor *)
{
    SoText2 *perfText = (SoText2 *) data;
     SbBool wasAnimating = FALSE;
     SbTime startTime;
     double lastFrames = 0;
    char str[100];

    assert(!useWalkViewer);

#ifdef _WIN32
    if (((SoWinExaminerViewer *)viewer)->isAnimating()) {
#else
    if (((SoQtExaminerViewer *)viewer)->isAnimating()) {
#endif
        if (wasAnimating) {
                SbTime timeDiff = SbTime::getTimeOfDay() - startTime;
                double frames =
                numFramesRendered * 1000.0 / timeDiff.getMsecValue();
            frames = floor(frames * 100.0 + 0.5) / 100.0;

            // Update the perf meter display if the numbers have changed
            if (frames != lastFrames) {
                sprintf (str,
                        "%d tris/frame,   %.2lf frames/sec,   %.0lf tris/sec",
                        numTris, frames, frames * numTris);
                    perfText->string.setValue(str);
                lastFrames = frames;
            }

            // Don't reset the timer here.  By using the total
            // time from when the object started animating
            // we will get more accurate numbers - especially
            // for slow objects.
        }
        else {
            // Reset the timer when we detect the object has
            // started animating.
            numFramesRendered = 0;
            startTime = SbTime::getTimeOfDay();
            wasAnimating = TRUE;
        }
    }
    else
        wasAnimating = FALSE;
}

//
// Reset the meter to only show the tris/sec number.
// Otherwise the numbers might look misleading
//
#ifdef _WIN32
 void
viewStartCB(void *data, SoWinViewer *)
#else
 void
viewStartCB(void *data, SoQtViewer *)
#endif
{
    char str[150];
    SoText2 *perfText = (SoText2 *) data;

    sprintf (str, "%d tris/frame", numTris);
    perfText->string.setValue(str);
}

////////////////////////////////////////////////////////////////////
//
// Turn busy cursor on/off
//
#if 0 // FIXME: Missing portable cursor functionality. PCJohn 2005-03-15
 void
setBusyCursor(SbBool showBusy)
{

    Display *display = shell ? XtDisplay(shell) : NULL;
    Window window = shell ? XtWindow(shell) : (Window) NULL;

    if (!shell || !display || !window) return;

    // make sure we have defined the cursor
    if (! busyCursor)
        createBusyCursor(display);

    if (showBusy)
        XDefineCursor(display, window, busyCursor);
    else
        XUndefineCursor(display, window);

    if (viewer)
        viewer->setCursorEnabled(!showBusy);

    // force the busy cursor to show up right now
    if (showBusy)
        XSync(display, False);
}
#endif
////////////////////////////////////////////////////////////////////
//
// Define busy cursor
//
#if 0 // FIXME: Missing portable cursor functionality. PCJohn 2005-03-15
 void
createBusyCursor(Display *display)
{
    Drawable d = DefaultRootWindow(display);
    XColor foreground, background;
    foreground.red = 65535;
    foreground.green = foreground.blue = 0;
    background.red = background.green = background.blue = 65535;

    Pixmap source = XCreateBitmapFromData(display, d,
        hourglass_bits, hourglass_width, hourglass_height);
    Pixmap mask = XCreateBitmapFromData(display, d,
        hourglass_mask_bits, hourglass_width, hourglass_height);

    busyCursor = XCreatePixmapCursor(display, source, mask,
        &foreground, &background, hourglass_x_hot, hourglass_y_hot);

    XFreePixmap(display, source);
    XFreePixmap(display, mask);
}
#endif

////////////////////////////////////////////////////////////////////
//
// Add a callback node to the main scene which will count the number
// of frames that have been rendered (to compute the tris/sec).
// Make sure the callback doesn't get cached or ruin the caching for
// the rest of the scene
//
 SoSeparator *
addUpdateCounter(SoNode *root)
{
    SoSeparator *newRoot = new SoSeparator;
    newRoot->renderCaching = SoSeparator::OFF;
    newRoot->boundingBoxCaching = SoSeparator::OFF;
    newRoot->pickCulling = SoSeparator::OFF;

    SoCallback *countUpdates = new SoCallback;
    countUpdates->setCallback(countUpdatesCB);

    newRoot->addChild(countUpdates);
    newRoot->addChild(root);

    return (newRoot);
}

 void
countUpdatesCB(void *, SoAction *action)
{
    if (action->isOfType(SoGLRenderAction::getClassTypeId())) {
        numFramesRendered++;
    }
}

///////////////////////////////////////////////////////////////////
//
// Brings up the "About..." dialog
//
 void
showAboutDialog()
{
#ifdef _WIN32
	OpenDokument("ivview.pdf");
#endif
#if 0 // Should be this ported?. PCJohn 2005-03-15

   if (access(IVPREFIX "/share/help/ivview/ivview.about", R_OK) != 0) {
        system("xmessage 'Sorry, could not find "
               IVPREFIX "/share/help/ivview/ivview.about' > /dev/null");
        return;
    }

    char command[100];
    sprintf(command, "which acroread > /dev/null");

    int err = system(command);
    if (err) {
        system("xmessage 'You must install acroread"
               " for this function to work' > /dev/null");
        return;
    }

    sprintf(command, "acroread " IVPREFIX "/share/help/ivview/ivview.about &");
    system(command);
#endif
}

///////////////////////////////////////////////////////////////////
//
// Use a motif file selection dialog to get the new filename.
// Set up a callback to open the file and create the new scenegraph.
//
 void
getNewScene()
{
#ifndef _WIN32

     Widget fileDialog = NULL;


    if (fileDialog == NULL) {
        Arg args[5];
        int n = 0;

        // Unmanage when ok/cancel are pressed
        XtSetArg(args[n], XmNautoUnmanage, TRUE); n++;
        fileDialog = XmCreateFileSelectionDialog(
            shell, "File Dialog", args, n);

        XtAddCallback(fileDialog, XmNokCallback,
                      (XtCallbackProc)newSceneCB, NULL);
    }

    // Manage the dialog
    XtManageChild(fileDialog);
#else
     TCHAR chFile[_MAX_PATH]; 
	 if (!Open_Dialog(chFile, NULL,"3D models (*.iv;*.wrl;*.vrml)\0*.iv;*.wrl;*.vrml\0All files(*.*)\0*.*\0"))
		    fprintf(stderr, "Error: open dialog\n");
		else {newSceneCB(chFile);}
#endif
}

//
// Callback routine that gets called when the new filename
// has been entered
//
#ifndef _WIN32
 void
newSceneCB(Widget, void *, XmFileSelectionBoxCallbackStruct *data)
{
	// Get the file name
    char *filename;
#else
 void
newSceneCB(TCHAR *filename)
{  
#endif  

	#ifndef _WIN32
	XmStringGetLtoR(data->value,
        (XmStringCharSet) XmSTRING_DEFAULT_CHARSET, &filename);
    #endif
    
    files.count = 1;
    files.names[0] = strdup(filename);
	#ifndef _WIN32
    XtFree(filename);
	#endif

    setBusyCursor(TRUE);
    SbBool readOk = readScene(files);
    setBusyCursor(FALSE);

    if (!readOk) {
        char str[100];
        strcpy(str, "Error reading file: ");
        strcat(str, files.names[0]);
		#ifndef _WIN32
        SoXt::createSimpleErrorDialog(shell, "File Error Dialog", str);
		#else
		MessageBox(NULL,"File Error Dialog","Error",MB_OK);
		#endif
        return;
    }
}


///////////////////////////////////////////////////////////////////
//
// Modify the viewer's background color using the color editor
// from the sample directory
//
 void
editBackgroundColor()
{
#if 0 // FIXME: This callback should be implemented when menu
      // functionality is ported. PCJohn 2005-03-15
     MyColorEditor *backgroundColorEditor = NULL;

    if (backgroundColorEditor == NULL) {
        backgroundColorEditor = new MyColorEditor;
        backgroundColorEditor->setColor(SbColor(0, 0, 0));
        backgroundColorEditor->setTitle( "Background Color" );
        backgroundColorEditor->addColorChangedCallback(
                                    backgroundColorCB);
    }

    backgroundColorEditor->show();
#endif
	// pridat otevreni vyberu barev
}

//
// Callback routine that gets called when the background
// color editor values change
//
 void
backgroundColorCB(void *, const SbColor *color)
{
    viewer->setBackgroundColor(*color);
}


///////////////////////////////////////////////////////////////////
// Create the viewer and lay it out so that it is attached
// to the bottom of the menu bar
//
#ifdef _WIN32
 SoWinViewer *
buildAndLayoutViewer(HWND parent, HWND menuWidget, SbBool walk)
#else
 SoQtViewer *
buildAndLayoutViewer(QWidget *parent, QWidget *menuWidget, SbBool walk)
#endif
{
#if 0 // FIXME: X-window creation code. This should be ported by using
      // Qt and WinAPI code. PCJohn 2005-03-15

    //
    //
    Arg args[12];
    int n = 0;
    SoXtViewer *viewer = NULL;
//    if (!walk) // Coin port
        viewer = new SoXtExaminerViewer(parent);
//    else
//        viewer = new SoXtWalkViewer(parent);

    viewer->setViewing(TRUE);        // default to VIEW

    XtSetArg(args[n], XmNtopAttachment,    XmATTACH_WIDGET); n++;
    XtSetArg(args[n], XmNtopWidget,        menuWidget); n++;
    XtSetArg(args[n], XmNbottomAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNleftAttachment,   XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNrightAttachment,  XmATTACH_FORM); n++;
    XtSetValues(viewer->getWidget(), args, n);

    return (viewer);
#else
    return NULL;
#endif
}



///////////////////////////////////////////////////////////////////
//

 void
countTriangleCB(void *userData, SoCallbackAction *,
                     const SoPrimitiveVertex *,
                const SoPrimitiveVertex *,
                const SoPrimitiveVertex *)
{
    int32_t *curCount = (int32_t *) userData;
    (*curCount)++;
}


int
countTriangles(SoNode *root)
{
    SoCallbackAction ca;
    int32_t numTris = 0;

    root->ref();

    ca.addTriangleCallback(SoShape::getClassTypeId(), countTriangleCB,
                           (void *) &numTris);
    ca.apply(root);

    root->unrefNoDelete();

    return (numTris);
}