/*---------------------------
Vvoj okennho systmu ve WinAPI
Bakalsk prce 2006, 
autor: Dolej Petr
login: xdolej06
Fakulta Informanch Technologi v Brn
---------------------------*/

#ifndef _WIN32_IVVIEW_FUN
#define _WIN32_IVVIEW_FUN

#include <Inventor/SoDB.h>
#include <Inventor/SoInput.h>
#include <Inventor/actions/SoCallbackAction.h>
#include <Inventor/actions/SoWriteAction.h>
#include <Inventor/elements/SoCacheElement.h>
#include <Inventor/nodes/SoAnnotation.h>
#include <Inventor/nodes/SoCallback.h>
#include <Inventor/nodes/SoFont.h>
#include <Inventor/nodes/SoOrthographicCamera.h>
#include <Inventor/nodes/SoShape.h>
#include <Inventor/nodes/SoSeparator.h>
#include <Inventor/nodes/SoText2.h>
#include <Inventor/sensors/SoTimerSensor.h>

# include <Inventor/Win/SoWin.h>
# include <Inventor/Win/viewers/SoWinExaminerViewer.h>
# include <Inventor/Win/devices/SoWinDevice.h>


#define SCREEN(w) XScreenNumberOfScreen(XtScreen(w))

typedef struct {
    int    count;        // Number of files read
    char **names;        // File names
    SoGroup *super;      // hidden top of the scene, added by PCJohn as workaround of overlays
    SoNode *original;        // Un-munged scene
    SoSeparator *optimized;  // Optimized scene
    SoSeparator *ivfixed;    // Ivfixed scene
} FileInfo;

// Function prototypes
 void print_usage();
 void parse_args(int , char **);
 SbBool readScene(FileInfo &);        // Returns TRUE if error
 void overlayViewportCB(void *, SoAction *);
 void countUpdatesCB(void *, SoAction *);
 void perfSensorCB(void *, SoSensor *);
 void showAboutDialog();
 void getNewScene();
 void optimizationChanged(FileInfo &, SbBool);
 void editBackgroundColor();
 void backgroundColorCB(void *, const SbColor *);
 SoSeparator *addUpdateCounter(SoNode *);
 int countTriangles(SoNode *);
 void countTriangleCB(void *, SoCallbackAction *,
                const SoPrimitiveVertex *, const SoPrimitiveVertex *,
                const SoPrimitiveVertex *);
#ifdef _WIN32
 void initPerfMeter(SoWinViewer *vwr);
 void viewStartCB(void *, SoWinViewer *);
 void newSceneCB(TCHAR *);
#else
 void initPerfMeter(SoQtViewer *vwr);
 void viewStartCB(void *, SoQtViewer *);
#endif

// Definitions for busy cursor:
//static Cursor busyCursor = 0;
#define hourglass_width 17
#define hourglass_height 21
#define hourglass_x_hot 8
#define hourglass_y_hot 11
static char hourglass_bits[] = {
   // char retyping done by PCJohn because of MSVC 6 warnings
   (char)0x00, (char)0x00, (char)0x00, (char)0xfe, (char)0xff, (char)0x00,
   (char)0x04, (char)0x40, (char)0x00, (char)0x04, (char)0x40, (char)0x00,
   (char)0x04, (char)0x40, (char)0x00, (char)0xe8, (char)0x2e, (char)0x00,
   (char)0xd0, (char)0x17, (char)0x00, (char)0xa0, (char)0x0b, (char)0x00,
   (char)0x40, (char)0x05, (char)0x00, (char)0x40, (char)0x05, (char)0x00,
   (char)0x40, (char)0x04, (char)0x00, (char)0x40, (char)0x04, (char)0x00,
   (char)0x40, (char)0x04, (char)0x00, (char)0x20, (char)0x09, (char)0x00,
   (char)0x10, (char)0x11, (char)0x00, (char)0x88, (char)0x23, (char)0x00,
   (char)0xc4, (char)0x47, (char)0x00, (char)0xe4, (char)0x4f, (char)0x00,
   (char)0xf4, (char)0x5f, (char)0x00, (char)0xfe, (char)0xff, (char)0x00,
   (char)0x00, (char)0x00, (char)0x00};

static char hourglass_mask_bits[] = {
   // char retyping done by PCJohn because of MSVC 6 warnings
   (char)0xff, (char)0xff, (char)0x01, (char)0xff, (char)0xff, (char)0x01,
   (char)0xfe, (char)0xff, (char)0x00, (char)0xfe, (char)0xff, (char)0x00,
   (char)0xfe, (char)0xff, (char)0x00, (char)0xfc, (char)0x7f, (char)0x00,
   (char)0xf8, (char)0x3f, (char)0x00, (char)0xf0, (char)0x1f, (char)0x00,
   (char)0xe0, (char)0x0f, (char)0x00, (char)0xe0, (char)0x0f, (char)0x00,
   (char)0xe0, (char)0x0f, (char)0x00, (char)0xe0, (char)0x0f, (char)0x00,
   (char)0xe0, (char)0x0f, (char)0x00, (char)0xf0, (char)0x1f, (char)0x00,
   (char)0xf8, (char)0x3f, (char)0x00, (char)0xfc, (char)0x7f, (char)0x00,
   (char)0xfe, (char)0xff, (char)0x00, (char)0xfe, (char)0xff, (char)0x00,
   (char)0xfe, (char)0xff, (char)0x00, (char)0xff, (char)0xff, (char)0x01,
   (char)0xff, (char)0xff, (char)0x01};

#endif